-- 嗨，你好嗎？
math["분야 명"] = math["値"]
--[[комментарий]] local t = {
   ["päällekkäinen nimi a​b"] = 1,
   ["päällekkäinen nimi a​b"] = 2
}

-- líne an-fhada a choinníonn dul ag dul agus ag dul agus ag dul agus ag dul ach nach bhfuil 120 carachtar ann. Y diwed
